__all__ = ['UserPassesTestMixin']


try:
    from django.contrib.auth.mixins import UserPassesTestMixin
except ImportError:  # pragma: no cover
    # Backport from Django 1.9.
    from django.conf import settings
    from django.contrib.auth.views import redirect_to_login
    from django.core.exceptions import ImproperlyConfigured, PermissionDenied
    from django.utils.encoding import force_text
    REDIRECT_FIELD_NAME = 'next'

    class AccessMixin(object):
        """
        Abstract CBV mixin that gives access mixins the same customizable
        functionality.
        """
        login_url = None
        permission_denied_message = ''
        raise_exception = False
        redirect_field_name = REDIRECT_FIELD_NAME

        def get_login_url(self):
            """
            Override this method to override the login_url attribute.
            """
            login_url = self.login_url or settings.LOGIN_URL
            if not login_url:
                raise ImproperlyConfigured(
                    '{0} is missing the login_url attribute. Define {0}.login_url, settings.LOGIN_URL, or override '
                    '{0}.get_login_url().'.format(self.__class__.__name__)
                )
            return force_text(login_url)

        def get_permission_denied_message(self):
            """
            Override this method to override the permission_denied_message attribute.
            """
            return self.permission_denied_message

        def get_redirect_field_name(self):
            """
            Override this method to override the redirect_field_name attribute.
            """
            return self.redirect_field_name

        def handle_no_permission(self):
            if self.raise_exception:
                raise PermissionDenied(self.get_permission_denied_message())
            return redirect_to_login(self.request.get_full_path(), self.get_login_url(),
                                     self.get_redirect_field_name())

    class UserPassesTestMixin(AccessMixin):
        """
        CBV Mixin that allows you to define a test function which must return True
        if the current user can access the view.
        """

        def test_func(self):
            raise NotImplementedError(
                '{0} is missing the implementation of the test_func() method.'.format(
                    self.__class__.__name__)
            )

        def get_test_func(self):
            """
            Override this method to use a different test_func method.
            """
            return self.test_func

        def dispatch(self, request, *args, **kwargs):
            user_test_result = self.get_test_func()()
            if not user_test_result:
                return self.handle_no_permission()
            return super(UserPassesTestMixin, self).dispatch(request, *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from django.core.management.base import BaseCommand

from quade.managers import manager


class Command(BaseCommand):

    def handle(self, *args, **kwargs):
        if len(manager.registry) == 0:
            self.stdout.write("0 functions are registered with Quade.")
        else:
            if len(manager.registry) == 1:
                opening = "1 function is"
            else:
                opening = "{} functions are".format(len(manager.registry))
            self.stdout.write("{} registered with Quade:".format(opening))
            for func_name in sorted(manager.registry):
                self.stdout.write("- {}".format(func_name))



from __future__ import absolute_import, division, print_function, unicode_literals

import importlib

from django.conf import settings
from django.db import transaction


class ConfigurationError(Exception):
    """
    A test configuration tried to execute a function that is not registered.
    """
    def __init__(self, unregistered_functions):
        self.unregistered_functions = unregistered_functions


class FixtureManager:
    def __init__(self):
        self._registry = {}

    def register(self, func):
        self._registry[func.__name__] = func

    @property
    def registry(self):
        return self._registry

    def setup(self):
        importlib.import_module(settings.QUADE.fixtures_file)

    def create(self, config):
        instructions = []
        for step in config:
            func_name, kwargs = step
            func = self._registry[func_name]
            instructions.append(func(**kwargs))
        return '\n'.join(instructions)

    def validate(self, config):
        """
        :return: raises or None
        """
        func_names = set(step[0] for step in config)
        registered_funcs = set(self._registry.keys())
        unregistered_funcs = func_names - registered_funcs
        if unregistered_funcs:
            raise ConfigurationError(unregistered_funcs)


manager = FixtureManager()  # Singleton pattern.


def register(func):
    """
    Register a function with the FixtureManager. Also wraps the function in an atomic transaction.
    """
    func = transaction.atomic(func)
    manager.register(func)
    return func



from __future__ import unicode_literals

from .settings import AllEnvs, DebugEnvs, Settings


__all__ = [
    'AllEnvs',
    'DebugEnvs',
    'Settings',
]


__version__ = '0.2.0'
default_app_config = 'quade.apps.QuadeConfig'



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-01-04 15:00
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_fsm.db.fields.fsmfield
import django_light_enums.db
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('instructions', models.TextField(blank=True, help_text='Information needed to run this particular scenario, such as usernames and login information.', null=True)),
                ('status', django_fsm.db.fields.fsmfield.FSMIntegerField(choices=[(1, b'READY'), (10, b'IN_PROGRESS'), (20, b'DONE'), (-1, b'NOT_READY'), (-10, b'FAILED')], db_index=True, default=-1)),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='RecordedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='recorded_objects', to='quade.Record')),
            ],
        ),
        migrations.CreateModel(
            name='Scenario',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(unique=True)),
                ('status', django_light_enums.db.EnumField(choices=[(0, b'INACTIVE'), (10, b'ACTIVE')], default=10, enum_values=[0, 10])),
                ('config', jsonfield.fields.JSONField()),
                ('description', models.TextField()),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('updated_on', models.DateTimeField(auto_now=True)),
            ],
            options={
                'ordering': ['description'],
            },
        ),
        migrations.AddField(
            model_name='record',
            name='scenario',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='quade.Scenario'),
        ),
    ]



from __future__ import absolute_import, division, print_function, unicode_literals

import functools
from django import forms
from django.conf import settings
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
try:
    from django.urls import reverse, reverse_lazy
except ImportError:
    from django.core.urlresolvers import reverse, reverse_lazy
from django.utils.functional import cached_property
from django.views.generic import View, FormView

from .compatability import UserPassesTestMixin
from .models import Record, Scenario
from .tasks import execute_test_task


class QuadeAccessMixin(UserPassesTestMixin):

    @cached_property
    def test_func(self):
        return functools.partial(settings.QUADE.access_test_func, self)

    raise_exception = True


class ExecuteTestForm(forms.Form):
    scenarios = forms.ModelChoiceField(
        queryset=Scenario.objects.none(),
        to_field_name='slug'
    )

    def __init__(self, *args, **kwargs):
        # Overwrite the queryset when the form is initialized to avoid stale data.
        super(ExecuteTestForm, self).__init__(*args, **kwargs)
        self.fields['scenarios'].queryset = Scenario.objects.active()

    def execute_test(self, created_by):
        scenario = self.cleaned_data['scenarios']
        record = Record.objects.create(scenario=scenario, created_by=created_by)
        if settings.QUADE.use_celery:
            execute_test_task.delay(record.id)
        else:
            record.execute_test()
        return record


class MainView(QuadeAccessMixin, FormView):
    template_name = 'quade/main.jinja'
    form_class = ExecuteTestForm
    success_url = reverse_lazy('quade-main')

    def get_context_data(self, **kwargs):
        context = super(MainView, self).get_context_data(**kwargs)
        # Remove the form from the context if there are no active tests currently.
        # (The default "-------" will always be present.)
        if self.request.method == 'GET' and len(context['form'].fields['scenarios'].choices) <= 1:
            context['form'] = None
        context['allowed'] = settings.QUADE.allowed
        context['recent_tests'] = Record.objects.all().order_by(
            '-created_on'
        ).select_related('scenario')[:30]
        return context

    def form_valid(self, form):
        form.execute_test(created_by=self.request.user)
        return super(MainView, self).form_valid(form)


class MarkDoneView(QuadeAccessMixin, View):
    def post(self, request, test_record_id):
        record = get_object_or_404(Record, id=test_record_id)
        record.status = Record.Status.DONE
        record.save()
        return HttpResponseRedirect(reverse('quade-main'))



from __future__ import absolute_import, division, print_function, unicode_literals

from celery import shared_task

from .models import Record


@shared_task
def execute_test_task(record_id):
    record = Record.objects.get(id=record_id)
    record.execute_test()



from __future__ import absolute_import, division, print_function, unicode_literals

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models as m
from django.utils.encoding import python_2_unicode_compatible
from django_fsm.db.fields import FSMIntegerField, transition
from django_light_enums import enum
from jsonfield import JSONField

from .managers import manager, ConfigurationError
from .receivers import connect_qa_object_receiver


class ScenarioManager(m.QuerySet):

    def active(self):
        return self.filter(status=Scenario.Status.ACTIVE)


@python_2_unicode_compatible
class Scenario(m.Model):
    """
    A specific scenario that QA tests will be run against.

    The heart of a ``Scenario`` is the `config` attribute. `config` is a list of several fixtures,
    possibly with arguments, which are executed sequentially to create and modify objects in the
    database.
    """

    class Meta:
        app_label = 'quade'
        ordering = ['description']

    class Status(enum.Enum):
        INACTIVE = 0
        ACTIVE = 10

    objects = ScenarioManager.as_manager()

    REGISTRY = {}

    slug = m.SlugField(unique=True)
    status = enum.EnumField(Status, default=Status.ACTIVE)
    config = JSONField()
    description = m.TextField()
    created_on = m.DateTimeField(auto_now_add=True)
    updated_on = m.DateTimeField(auto_now=True)

    def __str__(self):
        return '#{}: {}'.format(self.id, self.description)

    def activate(self):
        self.status = self.Status.ACTIVE
        self.save()

    def deactivate(self):
        self.status = self.Status.INACTIVE
        self.save()

    def delete(self):
        self.deactivate()

    def hard_delete(self, *args, **kwargs):
        super(Scenario, self).delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        try:
            manager.validate(self.config)
        except ConfigurationError as exc:
            raise ValidationError("config {} contains unregistered function(s): {}".format(
                self.config, ','.join(exc.unregistered_functions)
            ))
        super(Scenario, self).save(*args, **kwargs)


class Record(m.Model):
    """
    A record of setting up, and possibly executing, a particular :class:`Scenario`.
    """

    class Meta:
        app_label = 'quade'

    class Status(enum.Enum):
        FAILED = -10
        NOT_READY = -1
        READY = 1  # Django-FSM doesn't work well with 0, because it evaluates as falsey.
        IN_PROGRESS = 10
        DONE = 20

    scenario = m.ForeignKey(Scenario, on_delete=m.PROTECT)
    instructions = m.TextField(
        blank=True,
        null=True,
        help_text="Information needed to run this particular scenario, such as usernames and login"
        " information."
    )
    status = FSMIntegerField(choices=Status.choices, default=Status.NOT_READY)
    created_by = m.ForeignKey(settings.AUTH_USER_MODEL, related_name='+', on_delete=m.PROTECT)
    created_on = m.DateTimeField(auto_now_add=True)
    updated_on = m.DateTimeField(auto_now=True)

    def execute_test(self):
        """
        Public method for attempting to execute a test scenario. Exceptions put the record in the
        FAILED state and re-raise.
        """
        try:
            self._execute()
        except Exception as ex:
            self._fail(exception=ex)
            raise

    @transition(status, source=Status.NOT_READY, target=Status.READY, save=True)
    def _execute(self):
        with connect_qa_object_receiver(self):
            instructions = manager.create(self.scenario.config)
        self.instructions = instructions

    @transition(status, source=Status.NOT_READY, target=Status.FAILED, save=True)
    def _fail(self, exception):
        self.instructions = repr(exception)


@python_2_unicode_compatible
class RecordedObject(m.Model):
    """
    A joiner table for creating a generic many-to-many relation between :class:`Records <Record>`
    and any other objects.
    """

    class Meta:
        app_label = 'quade'

    content_type = m.ForeignKey(ContentType, on_delete=m.CASCADE)
    object_id = m.PositiveIntegerField()
    object = GenericForeignKey('content_type', 'object_id')
    record = m.ForeignKey(Record, related_name='recorded_objects', on_delete=m.CASCADE)

    def __str__(self):
        return "RecordedObject #{}: {}".format(self.pk, self.object)



# This module is inspired by and indebted to Hypothesis' _settings module.
# https://github.com/HypothesisWorks/hypothesis-python/blob/master/src/hypothesis/_settings.py
from collections import Iterable
from inspect import cleandoc

from attr import attrib, attrs
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from six import with_metaclass


__all__ = [
    'AllEnvs',
    'DebugEnvs',
    'Settings',
    'is_superuser',
]


@attrs
class Setting(object):
    """A descriptor that prevents deletion and provides a docstring."""
    name = attrib()
    default = attrib()
    default_description = attrib(default='')
    description = attrib(default='')
    validator = attrib(default=None)

    def __get__(self, obj, objtype=None):
        if obj is None:  # pragma: no cover
            return self
        else:
            try:
                return obj.__dict__[self.name]
            except KeyError:  # pragma: no cover
                raise AttributeError(self.name)

    def __set__(self, obj, value):
        obj.__dict__[self.name] = value

    def __delete__(self, obj):
        raise AttributeError('Cannot delete attribute {}'.format(self.name))

    @property
    def __doc__(self):
        return "{}\n\nDefault: {}".format(
            cleandoc(self.description),
            self.default_description or '``{}``'.format(self.default)
        )


class AllEnvs(object):
    """Enable Quade unconditionally."""
    @staticmethod
    def func(_): return True


class DebugEnvs(object):
    """Enable Quade if the environment has DEBUG = True."""
    @staticmethod
    def func(s): return s.DEBUG


def is_superuser(view):
    """Allow access to the view if the requesting user is a superuser."""
    return view.request.user.is_superuser


all_settings = {}


def validate_allowed_envs(val):
    if any([val in [AllEnvs, DebugEnvs], callable(val), isinstance(val, Iterable)]):
        return val
    else:
        raise TypeError


def validate_access_test_func(val):
    if callable(val):
        return val
    else:
        raise TypeError


def validate_boolean(val):
    if isinstance(val, bool):
        return val
    else:
        raise TypeError


class SettingsMeta(type):

    def __new__(cls, name, bases, dct):
        obj = super(SettingsMeta, cls).__new__(cls, name, bases, dct)
        obj.define_setting(
            name='fixtures_file',
            default='quade.fixtures',
            description="""A dotted path to the location of your fixtures.""",
        )
        obj.define_setting(
            name='allowed_envs',
            default=DebugEnvs,
            default_description=""":class:`.DebugEnvs`""",
            description="""Controls which environments Quade can run on.

            Allowed values are:

            - :class:`.DebugEnvs`

            - :class:`.AllEnvs`

            - a string -- your ``django.conf.settings.ENV`` is compared exactly to this string

            - an iterable -- your ``django.conf.settings.ENV`` is checked exactly for membership in
              the iterable

            - any callable that accepts one argument, the Django settings object

            If Quade is not enabled, fixtures will not be loaded, but views will be available to
            users with appropriate access, application models and tables can be accessed, etc.
            """,
            validator=validate_allowed_envs,
        )
        obj.define_setting(
            name='access_test_func',
            default=is_superuser,
            default_description=""":func:`.is_superuser`""",
            description="""A callable that restricts access to Quade's views. The callable should
            accept one argument (a Django view class).""",
            validator=validate_access_test_func
        )
        obj.define_setting(
            name='use_celery',
            default=False,
            description="""Whether to use Celery to set up a Record.

            Enabling this is useful if your fixtures take a long time to run.

            See :doc:`celery` for additional details.
            """,
            validator=validate_boolean,
        )
        return obj


class Settings(with_metaclass(SettingsMeta)):

    _WHITELISTED_PROPERTIES = ['_construction_complete']

    def __init__(self, **kwargs):
        self._construction_complete = False
        for setting in all_settings.values():
            val = kwargs.pop(setting.name, setting.default)
            if setting.validator:
                try:
                    val = setting.validator(val)
                except:
                    raise ImproperlyConfigured(
                        "{} is not a valid value for Quade setting {}".format(val, setting.name)
                    )
            setattr(self, setting.name, val)
        if kwargs:
            suffix = '' if len(kwargs) == 1 else 's'
            raise AttributeError("No such setting{}: '{}'".format(suffix, "', '".join(kwargs.keys())))
        self._construction_complete = True

    @classmethod
    def define_setting(cls, name, **kwargs):
        s = Setting(name, **kwargs)
        all_settings[name] = s
        setattr(cls, name, s)

    @property
    def allowed(self):
        """Based on the settings, determine if scenarios are allowed to run on this environment."""
        if self.allowed_envs in [AllEnvs, DebugEnvs]:
            func = self.allowed_envs.func
        elif callable(self.allowed_envs):
            func = self.allowed_envs
        elif isinstance(self.allowed_envs, six.string_types):
            def func(s): return s.ENV == self.allowed_envs
        elif isinstance(self.allowed_envs, Iterable):
            def func(s): return s.ENV in self.allowed_envs

        if func(settings):
            return True
        else:
            return False

    def __setattr__(self, key, value):
        if key in self._WHITELISTED_PROPERTIES:
            return super(Settings, self).__setattr__(key, value)
        elif key in all_settings:
            if self._construction_complete:
                raise RuntimeError('Quade settings may not be modified during runtime.')
            else:
                return super(Settings, self).__setattr__(key, value)
        else:  # pragma: no cover
            raise AttributeError('No such setting {}'.format(key))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

from setuptools import find_packages, setup


def get_version(*file_paths):
    """Retrieves the version."""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("src", "quade", "__init__.py")


if sys.argv[-1] == 'publish':
    try:
        import wheel
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    try:
        import twine
    except ImportError:
        print('Twine library missing. Please run "pip install twine"')
        sys.exit()
    os.system('python setup.py sdist bdist_wheel --universal')
    os.system('twine upload dist/*')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
# Replace relative internal reference to logo with an absolute external URL.
readme = readme.replace(
    'docs/_static/quade_200x200.png',
    'https://raw.githubusercontent.com/makingspace/quade/bf261f689b419f89b2125ccb9bd5bf01340f08e0/docs/_static/quade_200x200.png'
)

history = open('HISTORY.rst').read().replace('.. :changelog:\n\n', '')

setup(
    name='quade',
    version=version,
    description="""The friendly QA tool for Django.""",
    long_description=readme + '\n\n' + history,
    author='Paul Baranay',
    author_email='pbaranay@makespace.com',
    url='https://github.com/makingspace/quade',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=["attrs", "django-fsm==1.6", "django-light-enums", "django_jinja", "jsonfield"],
    extras_require={
        "celery": "celery"
    },
    license="MIT",
    zip_safe=False,
    keywords='django',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from __future__ import absolute_import, division, print_function, unicode_literals

from django.apps import AppConfig
from django.conf import settings

from .managers import manager


class QuadeConfig(AppConfig):
    name = 'quade'

    def ready(self):
        if settings.QUADE.allowed:
            manager.setup()



from __future__ import absolute_import, division, print_function, unicode_literals

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^$', views.MainView.as_view(), name='quade-main'),
    url(
        r'^record/(?P<test_record_id>[0-9]+)/done/$',
        views.MarkDoneView.as_view(),
        name='quade-mark-done'
    ),
]



from __future__ import absolute_import, division, print_function, unicode_literals

from contextlib import contextmanager
from functools import partial

from django.db.models.signals import post_save


def associate_instance_with_qa_record(record, instance, created, **kwargs):
    """
    A signal receiver that's meant to catch a post-save signal and associate a created model
    instance with the Record that generated it. Can only be used as a signal when the `record`
    argument is fixed (via functools.partial).
    """
    from .models import RecordedObject
    if created and not isinstance(instance, RecordedObject):  # Avoid infinite recursion
        RecordedObject.objects.create(record=record, object=instance)


@contextmanager
def connect_qa_object_receiver(record):
    """
    A context manager that connects the post-save signal to the associate_instance_with_qa_record
    receiver, and ensures the receiver is disconnected on exit.
    """
    callback = partial(associate_instance_with_qa_record, record=record)
    post_save.connect(callback)
    yield callback
    post_save.disconnect(callback)
